/*   File: AI_6.cc
     Computes the integrals ...
     

     Author:  Tomas Johnson <johnson@math.uu.se>
     Latest edit: fre aug 29 14:17:49 CEST 2008

     By: Tomas Johnson <johnson@math.uu.se>
*/

#include <fstream>
#include <cstdlib>
#include <iostream>
#include <string>
#include <sstream>
#include <stdexcept>
#include <time.h>
#include "imath.hpp"
#include "interval.hpp"
#include "ivector.hpp"
#include "imatrix.hpp"
#include "matinv.hpp"

#include "list.h"

using namespace cxsc;
using namespace std;

//Performance variables:
int intYes(0);
int intNo(0);

//Global constants, initiated at run-time:
int oval;
int d;

//The constants in the Hamiltonian
const interval a_20=interval(1.0)/2.0;
const interval a_40=interval(-9.0)/8.0;
const interval a_60=interval(1.0)/3.0;
const interval a_02=interval(1.0)/2.0;
const interval a_04=interval(-73.0)/144.0;
const interval a_06=interval(2.0)/27.0;

bool back(false);  //if running backwards (in h)

enum command {START_TIMING, SHOW_TIMING, STOP_TIMING};
////////////////////////////////////////////////////////////////////////////
//
// 'START_TIMING' starts the clock, 'STOP_TIMING' stops the clock,
// and prints the elapsed time in hh:mm:ss. 'SHOW_TIMING' prints the
// elapsed time.
//
void clock(const command &cmd)
{
  struct tm     *ptr;          // Used for timing the program
  static time_t  time_start;   // start time

  if ( cmd == START_TIMING ) {
    time_start = time(NULL);
    ptr = localtime(&time_start);
    cout << "Started computations: " << asctime(ptr);
  }
  else if ( cmd == SHOW_TIMING || cmd == STOP_TIMING ) {
    int weeks   = 0;
    int days    = 0;
    int hours   = 0;
    int minutes = 0;
    double seconds;
    time_t time_stop;

    time_stop = time(NULL);
    ptr = localtime(&time_stop);
    if ( cmd == STOP_TIMING )
      cout << "Completed computations: " << asctime(ptr);
    seconds = difftime(time_stop, time_start);
    cout << "Elapsed run time: ";
    // Convert seconds to vv:dd:hh:mm:ss
    while ( seconds >= 60 ) { seconds -= 60; minutes++; }
    while ( minutes >= 60 ) { minutes -= 60;   hours++; }
    while (   hours >= 24 ) {   hours -= 24;    days++; }
    while (    days >=  7 ) {    days -=  7;   weeks++; }
    if ( weeks > 0)
      cout << weeks << " week" << (weeks != 1 ? "s, " : ", ");
    if ( days > 0)
      cout << days << " day" << (days != 1 ? "s, " : ", ");
    if ( hours > 0)
      cout << hours << " hour" << (hours != 1 ? "s, " : ", ");
    if ( minutes > 0)
      cout << minutes << " minute" << (minutes != 1 ? "s, " : ", ");
    cout << seconds << " second" << (seconds != 1 ? "s." : ".") << endl;
  }
}

///////////////////////////////////////////////////////////////////////////
//
static string int2str(int number)
{
   ostringstream oss;
   oss << number;
   return oss.str();
}


////////////////////////////////////////////////////////////////////////////

bool Intersect(interval &res, const interval &in1, const interval &in2) {
  if(Sup(in1)<Inf(in2) || Sup(in2)<Inf(in1))
    return false;
  res=in1&in2;
  return true;
}

bool Intersect(ivector &res, const ivector &in1, const ivector &in2) {
  int n=Ub(in1);
  if(n!=Ub(in2))
    return false;
  for(int i=1; i<=n; i++) {
    if (!Intersect(res[i], in1[i], in2[i]) )
      return false;
  }
  return true;
}

//Equivariant (pi) ORDER 6
ivector gradH(const ivector &Box) {
  ivector res(2);
  res[1]=Box[1]*(1.0+sqr(Box[1])*(4.0*a_40+6.0*a_60*sqr(Box[1])));
  res[2]=Box[2]*(1.0+sqr(Box[2])*(4.0*a_04+6.0*a_06*sqr(Box[2])));
  if(!back)
    return res;
  return -res;
}

rvector gradH(const rvector &onOval) {
  return mid(gradH(ivector(onOval)));
}

interval H(const ivector &Box) {
  ivector grad=gradH(Box);
  interval xprim(grad[1]);
  interval yprim(grad[2]);
  int xMono=in(0.0, xprim);
  int yMono=in(0.0, yprim);
 
  interval res=sqr(Box[1])*(a_20+sqr(Box[1])*(a_40+a_60*sqr(Box[1])))
    +sqr(Box[2])*(a_02+sqr(Box[2])*(a_04+a_06*sqr(Box[2])));
  
  if (xMono==0 && yMono==0)
    res=res&(sqr(Inf(Box[1]))*(a_20+sqr(Inf(Box[1]))*(a_40+a_60*sqr(Inf(Box[1]))))
	     +sqr(Inf(Box[2]))*(a_02+sqr(Inf(Box[2]))*(a_04+a_06*sqr(Inf(Box[2])))) |
	     
	     sqr(Inf(Box[1]))*(a_20+sqr(Inf(Box[1]))*(a_40+a_60*sqr(Inf(Box[1]))))
	     +sqr(Sup(Box[2]))*(a_02+sqr(Sup(Box[2]))*(a_04+a_06*sqr(Sup(Box[2])))) |
	     
	     sqr(Sup(Box[1]))*(a_20+sqr(Sup(Box[1]))*(a_40+a_60*sqr(Sup(Box[1]))))
	     +sqr(Inf(Box[2]))*(a_02+sqr(Inf(Box[2]))*(a_04+a_06*sqr(Inf(Box[2])))) |
	     
	     sqr(Sup(Box[1]))*(a_20+sqr(Sup(Box[1]))*(a_40+a_60*sqr(Sup(Box[1]))))
	     +sqr(Sup(Box[2]))*(a_02+sqr(Sup(Box[2]))*(a_04+a_06*sqr(Sup(Box[2])))) ); 
  
  else if (xMono==0)
    res=res&( sqr(Inf(Box[1]))*(a_20+sqr(Inf(Box[1]))*(a_40+a_60*sqr(Inf(Box[1]))))
	      +sqr(Box[2])*(a_02+sqr(Box[2])*(a_04+a_06*sqr(Box[2]))) |
	      
	      sqr(Sup(Box[1]))*(a_20+sqr(Sup(Box[1]))*(a_40+a_60*sqr(Sup(Box[1]))))
	      +sqr(Box[2])*(a_02+sqr(Box[2])*(a_04+a_06*sqr(Box[2]))) );
  
  else if (yMono==0)
    res=res&(sqr(Box[1])*(a_20+sqr(Box[1])*(a_40+a_60*sqr(Box[1])))
	     +sqr(Inf(Box[2]))*(a_02+sqr(Inf(Box[2]))*(a_04+a_06*sqr(Inf(Box[2])))) |
	     
	     sqr(Box[1])*(a_20+sqr(Box[1])*(a_40+a_60*sqr(Box[1])))
	     +sqr(Sup(Box[2]))*(a_02+sqr(Sup(Box[2]))*(a_04+a_06*sqr(Sup(Box[2])))) );
  
  interval HMid=sqr(mid(Box[1]))*(a_20+sqr(mid(Box[1]))*(a_40+a_60*sqr(mid(Box[1]))))
    +sqr(mid(Box[2]))*(a_02+sqr(mid(Box[2]))*(a_04+a_06*sqr(mid(Box[2]))));
  
  res=res&(HMid+xprim*(Box[1]-mid(Box[1]))+yprim*(Box[2]-mid(Box[2])));
  
  if(!back)
    return res;
  return -res;
}

void Integrate(const int &i, imatrix &Imono, const ivector Box) {
  Imono[i][1]+=interval(diam(Box[1])*diam(Box[2]));
  Imono[i][2]+=interval(power(Sup(Box[1]),3)-power(Inf(Box[1]),3))/3.0*diam(Box[2]);
  Imono[i][3]+=interval(power(Sup(Box[2]),3)-power(Inf(Box[2]),3))/3.0*diam(Box[1]);
  Imono[i][4]+=interval(power(Sup(Box[1]),5)-power(Inf(Box[1]),5))/5.0*diam(Box[2]);
  Imono[i][5]+=interval(power(Sup(Box[1]),3)-power(Inf(Box[1]),3))/3.0
    *interval(power(Sup(Box[2]),3)-power(Inf(Box[2]),3))/3.0;
  Imono[i][6]+=interval(power(Sup(Box[2]),5)-power(Inf(Box[2]),5))/5.0*diam(Box[1]);
}

void boundaryIntegrate(const int &i, imatrix &Imono, const ivector &Box) {
  Imono[i][1]+=interval(0.0, diam(Box[1])*diam(Box[2]) );
  Imono[i][2]+=interval(0.0)|interval(power(Sup(Box[1]),3)-power(Inf(Box[1]),3))/3.0*diam(Box[2]);
  Imono[i][3]+=interval(0.0)|interval(power(Sup(Box[2]),3)-power(Inf(Box[2]),3))/3.0*diam(Box[1]);
  Imono[i][4]+=interval(0.0)|interval(power(Sup(Box[1]),5)-power(Inf(Box[1]),5))/5.0*diam(Box[2]);
  Imono[i][5]+=interval(0.0)|(interval(power(Sup(Box[1]),3)-power(Inf(Box[1]),3))/3.0
			      *interval(power(Sup(Box[2]),3)-power(Inf(Box[2]),3))/3.0);
  Imono[i][6]+=interval(0.0)|interval(power(Sup(Box[2]),5)-power(Inf(Box[2]),5))/5.0*diam(Box[1]);
}

bool intersectOval(const interval &k, ivector &left, 
		      const interval &indep, const int &indepIndex, const interval &h) {
  // cout<<indep<<endl;
  bool value(false), diff(false);
  if(indepIndex==1) {
    for(int i=0; i<=199; i++){
      ivector cand(2);
      cand[1]=Inf(indep)+interval(i, i+1.0)/200.0*interval(Sup(indep)-Inf(indep));
      //  cout<<i<<" "<<cand[1];
      cand[2]=left[2]+k*(cand[1]-left[1]);
      ivector grad=gradH(cand);
      interval Hprim=grad[1]+k*grad[2];   
      interval temp;
      if( in(0.0,Hprim) )
	diff=true;
      if(Intersect(temp, h, H(cand)) )
	value=true;
    }
  }
  else {
    for(int i=0; i<=199; i++){
      ivector cand(2);
      cand[2]=Inf(indep)+interval(i, i+1.0)/200.0*interval(Sup(indep)-Inf(indep));
      cand[1]=left[1]+k*(cand[2]-left[2]);
      ivector grad=gradH(cand);
      interval Hprim=k*grad[1]+grad[2];   
      interval temp;
      if( in(0.0,Hprim) )
	diff=true;
      if(Intersect(temp, h, H(cand)) )
	value=true;
    }
  }
  //  cout<<endl;
  return value&&diff;
}

bool ovalBoundaryIntersect(const interval &h, const ivector &Box, const int &Side) {
  ivector cand(2);
  ivector encl(2);
  int n;
  if(Side==1) {
    cand[1]=interval(Sup(Box[1]));
    encl[1]=interval(Sup(Box[1]));
    n=2;
  }
  else if(Side==2) {
    cand[2]=interval(Sup(Box[2]));
    encl[2]=interval(Sup(Box[2]));
    n=1;
  }
  else if(Side==3) {
   cand[1]=interval(Inf(Box[1]));
   encl[1]=interval(Inf(Box[1]));
   n=2;
  }
  else if(Side==4) {
    cand[2]=interval(Inf(Box[2]));
    encl[2]=interval(Inf(Box[2]));
    n=1;
  }

  cand[n]=interval(mid(Box[n]));
  encl[n]=Box[n];
  
  bool intersect(true);
  
  for(int i=1; i<=8; i++) {
    interval Hmh=H(cand)-h;
    interval Hmhprim=(gradH(encl))[n];
    if(in(0, Hmhprim))
      return true;
    
    interval N=cand[n]-Hmh/Hmhprim;
    if(!Intersect(encl[n],encl[n],N)) {
      i=10;
      intersect=false;
    }
    else
      cand[n]=interval(mid(encl[n]));
  }
 
  return intersect;
}

bool zeroHOfY(const interval &h, 
	      const real &candX, real &candY, interval searchY){
  ivector cand(2);
  cand[1]=interval(candX);
  cand[2]=interval(candY);
  ivector encl(2);
  encl[1]=interval(candX);
  bool unique(false);
  for(int i=1; i<=5; i++) {
    encl[2]=searchY;
    interval Hmh=H(cand)-h;
    interval Hmhprim=(gradH(encl))[2];
    if(in(0, Hmhprim)) {
      candY=mid(cand[2]);
      return false;
    }
    interval N=cand[2]-Hmh/Hmhprim;
    if(in(N,searchY))
      unique=true;
    if(!Intersect(searchY,searchY,N)) {
      candY=mid(cand[2]);
      return false;
    }
    cand[2]=interval(mid(searchY));
  }
  candY=mid(cand[2]);
  return unique;
}

bool zeroHOfX(const interval &h, 
	      real &candX, const real &candY, interval searchX){
  ivector cand(2);
  cand[1]=interval(candX);
  cand[2]=interval(candY);
  ivector encl(2);
  encl[2]=interval(candY);
  bool unique(false);
  for(int i=1; i<=5; i++) {
    encl[1]=searchX;
    interval Hmh=H(cand)-h;
    interval Hmhprim=(gradH(encl))[1];
    if(in(0, Hmhprim)) {
      candX=mid(cand[1]);
      return false;
    }
    interval N=cand[1]-Hmh/Hmhprim;
    if(in(N,searchX))
      unique=true;
    if(!Intersect(searchX,searchX,N)) {
      candX=mid(cand[1]);
      return false;
    }
    cand[1]=interval(mid(searchX));
  }
  candX=mid(cand[1]);
  return unique;
}


bool triangle(const int &Index, const ivector &C1, const ivector &C2, const ivector &C3, 
	      const interval &h, const int &i, imatrix &Imono) {
  interval triArea;
  ivector triEnclosure;
  triEnclosure=C1|C2|C3;

  if(Index==1){
    ivector testPt=(C1+C2+C3)/3;
    interval Htest=H(testPt);
    if(Inf(Htest)>Sup(h))
      return true;
    else if(Sup(Htest)<Inf(h)) {
      interval x0=interval(Inf(C3[1]));
      interval x1=interval(Sup(C1[1]));
      interval y0=interval(Inf(C1[2]));
      interval y1=interval(Sup(C2[2]));
      if(x1==x0 || y0==y1)
	return true;

      triArea=(y1-y0)*(x1-x0)/2.0;
    }
    else {cout<<"err"<<endl;
      return false; }
    
  }
  else if(Index==2) {
    ivector testPt=(C1+C2+C3)/3;
    interval Htest=H(testPt);
    if(Inf(Htest)>Sup(h))
      return true;
    else if(Sup(Htest)<Inf(h)) {
      interval x0=interval(Inf(C2[1]));
      interval x1=interval(Sup(C1[1]));
      interval y0=interval(Inf(C3[2]));
      interval y1=interval(Sup(C1[2]));
      if(x1==x0 || y0==y1)
	return true;

      triArea=(y1-y0)*(x1-x0)/2.0;
    }
    else {cout<<"err"<<endl;
      return false;}
  }
  else if(Index==3){
    ivector testPt=(C1+C2+C3)/3;
    interval Htest=H(testPt);
    if(Inf(Htest)>Sup(h))
      return true;
    else if(Sup(Htest)<Inf(h)) {
      interval x0=interval(Inf(C2[1]));
      interval x1=interval(Sup(C1[1]));
      interval y0=interval(Inf(C3[2]));
      interval y1=interval(Sup(C2[2]));
      if(x1==x0 || y0==y1)
	return true;

      triArea=(y1-y0)*(x1-x0)/2.0;
    }
    else {cout<<"err"<<endl;
      return false;}
  }
  else{  //Index==4
    ivector testPt=(C1+C2+C3)/3;
    interval Htest=H(testPt);
    if(Inf(Htest)>Sup(h))
      return true;
    else if(Sup(Htest)<Inf(h)) {
      interval x0=interval(Inf(C3[1]));
      interval x1=interval(Sup(C1[1]));
      interval y0=interval(Inf(C3[2]));
      interval y1=interval(Sup(C2[2]));
      if(x1==x0 || y0==y1)
	return true;

      triArea=(y1-y0)*(x1-x0)/2.0;
    }
    else {cout<<"err"<<endl;
      return false;}
  }

  triArea=abs(triArea);

  Imono[i][1]+=triArea;
  Imono[i][2]+=triArea*sqr(triEnclosure[1]);
  Imono[i][3]+=triArea*sqr(triEnclosure[2]);
  Imono[i][4]+=triArea*power(triEnclosure[1],4);
  Imono[i][5]+=triArea*sqr(triEnclosure[1])*sqr(triEnclosure[2]);
  Imono[i][6]+=triArea*power(triEnclosure[2],4);

  return true;

}


bool boundaryIntegrateRot(const int &i, interval h, imatrix &Imono, const ivector &Box, const real &tol) {
 
  rvector onOval=mid(Box);
  rvector grad=gradH( onOval);
  if(abs(grad[1])>abs(grad[2]))
    zeroHOfX(h, onOval[1], onOval[2], Box[1]);
  else
    zeroHOfY(h,onOval[1], onOval[2], Box[2]);

  grad=gradH( onOval);              
 
  rvector appOval(2);
  appOval[1]=-grad[2];
  appOval[2]=grad[1];

  if(appOval[1]<0.0)
    appOval=-appOval;
  real rightX, leftX, rightY, leftY;
  int right, left;

  real gradTol=power(10,-14);
  if(appOval[2]>0.0) {
    if(abs(appOval[2])< gradTol) {
      rightX=Sup(Box[1]);
      leftX=Inf(Box[1]);
      right=1;
      left=3;
      rightY=onOval[2];
      leftY=onOval[2];
    } 
    else if(abs(appOval[1])< gradTol) {
      rightX=onOval[1];
      leftX=onOval[1];
      right=2;
      left=4;
      rightY=Sup(Box[2]);
      leftY=Inf(Box[2]);      
    }
    else {
      rightX=min(Sup(Box[1]), onOval[1]+(Sup(Box[2])-onOval[2])*appOval[1]/appOval[2]);
      if(rightX==Sup(Box[1]))
	right=1;
      else
	right=2;
      leftX =max(Inf(Box[1]), onOval[1]+(Inf(Box[2])-onOval[2])*appOval[1]/appOval[2]);
    
      if(leftX==Inf(Box[1]))
	left=3;
      else
	left=4;
      
      rightY=min(Sup(Box[2]), onOval[2]+(Sup(Box[1])-onOval[1])*appOval[2]/appOval[1]);
      leftY =max(Inf(Box[2]), onOval[2]+(Inf(Box[1])-onOval[1])*appOval[2]/appOval[1]);
    }
  }  
  else {
    if(abs(appOval[2])< gradTol) {
      rightX=Sup(Box[1]);
      leftX=Inf(Box[1]);
      right=1;
      left=3;
      rightY=onOval[2];
      leftY=onOval[2];
    }
    else if(abs(appOval[1])< gradTol) {
      rightX=onOval[1];
      leftX=onOval[1];
      right=4;
      left=2;
      rightY=Inf(Box[2]);
      leftY=Sup(Box[2]);     
    }
    else {
      rightX=min(Sup(Box[1]), onOval[1]+(Inf(Box[2])-onOval[2])*appOval[1]/appOval[2]);
      if(rightX==Sup(Box[1]))
	right=1;
      else
	right=4;
      leftX =max(Inf(Box[1]), onOval[1]+(Sup(Box[2])-onOval[2])*appOval[1]/appOval[2]);
      if(leftX==Inf(Box[1]))
	left=3;
      else
	left=2;
   
      rightY=max(Inf(Box[2]), onOval[2]+(Sup(Box[1])-onOval[1])*appOval[2]/appOval[1]);
      leftY =min(Sup(Box[2]), onOval[2]+(Inf(Box[1])-onOval[1])*appOval[2]/appOval[1]);
    }
  }
  
  //Now locate a zero of H-h on the intersecting sides !
  //Then we
  //Must prove that the lines from rightUp to leftUp and viceversa 
  // are outside of the spectrum !
  ivector rightUp(2), leftUp(2), rightDown(2), leftDown(2);
  if (right==1) {
    if(!zeroHOfY(h, rightX, rightY, Box[2]))
      return false;
    rightUp[1]=interval(Sup(Box[1]));
    rightUp[2]=interval(min(rightY+tol, Sup(Box[2])));
    rightDown[1]=interval(Sup(Box[1]));
    rightDown[2]=interval(max(rightY-tol, Inf(Box[2]))); 
  }
  else if(right==2) {
    if(!zeroHOfX(h, rightX, rightY, Box[1]))
      return false;
    rightUp[1]=interval(max(rightX-tol, Inf(Box[1])));
    rightUp[2]=interval(Sup(Box[2]));
    rightDown[1]=interval(min(rightX+tol,Sup(Box[1])));
    rightDown[2]=interval(Sup(Box[2]));      
  }
  else { //4
    if(!zeroHOfX(h, rightX, rightY, Box[1]))
      return false;
    rightUp[1]=interval(min(rightX+tol, Sup(Box[1])));
    rightUp[2]=interval(Inf(Box[2]));
    rightDown[1]=interval(max(rightX-tol, Inf(Box[1])));
    rightDown[2]=interval(Inf(Box[2]));   
  }
  
  if (left==3) {
    if(!zeroHOfY(h, leftX, leftY, Box[2]))
      return false;
    leftUp[1]=interval(Inf(Box[1]));
    leftUp[2]=interval(min(leftY+tol, Sup(Box[2])));
    leftDown[1]=interval(Inf(Box[1]));
    leftDown[2]=interval(max(leftY-tol, Inf(Box[2])));      
  }
  else if(left==2) {
    if(!zeroHOfX(h, leftX, leftY, Box[1]))
      return false;
    leftUp[1]=interval(min(leftX+tol, Sup(Box[1])));
    leftUp[2]=interval(Sup(Box[2]));
    leftDown[1]=interval(max(leftX-tol, Inf(Box[1])));
    leftDown[2]=interval(Sup(Box[2]));      
  }
  else { //4
    if(!zeroHOfX(h, leftX, leftY, Box[1]))
      return false;
    leftUp[1]=interval(max(leftX-tol, Inf(Box[1])));
    leftUp[2]=interval(Inf(Box[2]));
    leftDown[1]=interval(min(leftX+tol, Sup(Box[1])));
    leftDown[2]=interval(Inf(Box[2]));      
  }

  int notSide1, notSide2;
  if(right==1 && left==2) {
    notSide1=3;
    notSide2=4;
  }
  else if(right==1 && left==3) {
    notSide1=2;
    notSide2=4;
  }
  else if(right==1 && left==4) {
    notSide1=2;
    notSide2=3;
  }
  else if(right==2 && left==3) {
    notSide1=1;
    notSide2=4;
  }
  else if((right==2 && left==4) || (right==4 && left==2) ){
    notSide1=1;
    notSide2=3;
  }
  else if(right==4 && left==3) {
    notSide1=1;
    notSide2=2;
  }

  if(ovalBoundaryIntersect( h, Box, notSide1) ||
     ovalBoundaryIntersect( h, Box, notSide2) ) {
    return false;
  }

  //Make sure that up and down are on different sides
  //Suffices to prove for one if the test below works.
  if( !(Sup((H( rightDown)-h)*(H( rightUp)-h)) < 0.0) )
    return false;

  //Make sure both up and down are on the same side
  if ( !( 0.0 < Inf((H( rightUp)-h)*(H(leftUp)-h)) 
	  &&  0.0 < Inf((H( rightDown)-h)*(H(leftDown)-h)) ) )
    return false;
 
  if(abs(grad[2])>abs(grad[1])) {
    if(in(0.0,rightUp[1]-leftUp[1]))
      return false;
   
    interval k=(rightUp[2]-leftUp[2])/(rightUp[1]-leftUp[1]);
    interval x=leftUp[1]|rightUp[1];
    
    if( intersectOval( k, leftUp,x,1,h) )
      return false;

    if(in(0.0,rightDown[1]-leftDown[1])) 
      return false;
   
    k=(rightDown[2]-leftDown[2])/(rightDown[1]-leftDown[1]);
    x=leftDown[1]|rightDown[1];
 
    if( intersectOval( k, leftDown,x,1,h) )
      return false; 
 
  }

  else {
    if(in(0.0,rightUp[2]-leftUp[2]))
      return false;
    
    interval k=(rightUp[1]-leftUp[1])/(rightUp[2]-leftUp[2]);
    interval y=leftUp[2]|rightUp[2];

    if( intersectOval( k, leftUp,y,2,h) )
      return false; 
    
    
    if(in(0.0,rightDown[2]-leftDown[2]))
      return false;
    
    k=(rightDown[1]-leftDown[1])/(rightDown[2]-leftDown[2]);
    y=leftDown[2]|rightDown[2];
  
    if( intersectOval( k, leftDown,y,2,h) )
      return false;
  
  }
  
  //Time to integrate !!!
    
  interval errArea;
  ivector Box1(2), Box2(2);
  if (right==1 && left==2) {
    errArea=((leftDown[1]-Sup(Box[1]))*(rightDown[2]-Sup(Box[2]))-(leftUp[1]-Sup(Box[1]))*(rightUp[2]-Sup(Box[2])))/2;
    if(!triangle(1, rightUp, ivector(Sup(Box)), leftUp,  h, i, Imono))
      return false;
    ivector temp(2);
    temp[1]=Inf(leftDown[1]);
    temp[2]=Inf(rightDown[2]);
    if(!triangle(3, rightDown, leftDown, temp,  h, i, Imono))
      return false;
 
    Box1[1]=interval(Inf(Box[1]), Inf(leftDown[1]));
    Box1[2]=Box[2];
    Box2[1]=interval(Inf(leftDown[1]), Sup(Box[1]));
    Box2[2]=interval(Inf(Box[2]), Inf(rightDown[2]));

  }
  else if (right==1 && left==3) {
    errArea=2*tol*diam(Box[1]); 

    if(Inf(rightUp[2])>Sup(leftUp[2])) {
      ivector temp(2);
      temp[1]=Inf(Box[1]);
      temp[2]=Sup(rightUp[2]);
      if(!triangle(2, rightUp, temp, leftUp,  h, i, Imono))
	return false;
    
      temp[1]=Sup(Box[1]);
      temp[2]=Inf(leftDown[2]);
      if(!triangle(4, temp, rightDown, leftDown,  h, i, Imono))
	return false;
    }
    else {
      ivector temp(2);
      temp[1]=Sup(Box[1]);
      temp[2]=Sup(leftUp[2]);
      if(!triangle(1, rightUp, temp, leftUp,  h, i, Imono))
	return false;

      temp[1]=Inf(Box[1]);
      temp[2]=Inf(rightDown[2]);
      if(!triangle(3, rightDown, leftDown, temp,  h, i, Imono))
	return false;
    }

    Box1[1]=Box[1];
    Box1[2]=interval(max(Sup(rightUp[2]),Sup(leftUp[2])), Sup(Box[2]));
    Box2[1]=Box[1];
    Box2[2]=interval(Inf(Box[2]), min(Inf(rightDown[2]), Inf(leftDown[2])));
  }
  else if (right==1 && left==4) {
    errArea=((leftUp[1]-Sup(Box[1]))*(rightUp[2]-Inf(Box[2]))-(leftDown[1]-Sup(Box[1]))*(rightDown[2]-Inf(Box[2])))/2;

    ivector temp(2);
    temp[1]=Inf(leftUp[1]);
    temp[2]=Sup(rightUp[2]);
    if(!triangle(2, rightUp, temp, leftUp,  h, i, Imono))
      return false;
    
    temp[1]=Sup(Box[1]);
    temp[2]=Inf(Box[2]);
    if(!triangle(4, temp, rightDown, leftDown,  h, i, Imono))
      return false;

    Box1[1]=interval(Inf(Box[1]), Inf(leftUp[1]));
    Box1[2]=interval(Inf(Box[2]), Sup(rightUp[2]));
    Box2[1]=Box[1];
    Box2[2]=interval(Sup(rightUp[2]), Sup(Box[2]));

  }
  else if(right==2 && left==3) {
    errArea=((leftDown[2]-Sup(Box[2]))*(rightDown[1]-Inf(Box[1]))-(leftUp[2]-Sup(Box[2]))*(rightUp[1]-Inf(Box[1])))/2; 
    ivector temp(2);
    temp[1]=Inf(Box[1]);
    temp[2]=Sup(Box[2]);
    if(!triangle(2, rightUp, temp, leftUp,  h, i, Imono))
      return false;
    
    temp[1]=Sup(rightDown[1]);
    temp[2]=Inf(leftDown[2]);
    if(!triangle(4, temp, rightDown, leftDown,  h, i, Imono))
      return false;
    
    Box1[1]=interval(Sup(rightDown[1]), Sup(Box[1]));
    Box1[2]=interval(Sup(leftUp[2]), Sup(Box[2]));
    Box2[1]=Box[1];
    Box2[2]=interval(Inf(Box[2]), Sup(leftUp[2]));

  }
  else if((right==2 && left==4) || (right==4 && left==2) ) {
    errArea=2*tol*diam(Box[2]);   

    if(right==2) {
      ivector temp(2);
      temp[1]=Inf(leftUp[1]);
      temp[2]=Sup(Box[2]);
      if(!triangle(2, rightUp, temp, leftUp,  h, i, Imono))
	return false;
      
      temp[1]=Sup(rightDown[1]);
      temp[2]=Inf(Box[2]);
      if(!triangle(4, temp, rightDown, leftDown,  h, i, Imono))
	return false;
    }
    else {
      ivector temp(2);
      temp[1]=Sup(rightUp[1]);
      temp[2]=Sup(Box[2]);
      if(!triangle(1, rightUp, temp, leftUp,  h, i, Imono))
	return false;
      
      temp[1]=Inf(leftDown[1]);
      temp[2]=Inf(Box[2]);
      if(!triangle(3, rightDown, leftDown, temp,  h, i, Imono))
	return false;
    }

    Box1[1]=interval(Inf(Box[1]), min(Inf(leftDown[1]), Inf(leftUp[1])));
    Box1[2]=Box[2];
    Box2[1]=interval(max(Sup(rightDown[1]), Sup(rightUp[1])), Sup(Box[1]));
    Box2[2]=Box[2];
    
  }
  else if(right==4 && left==3) {
    errArea=((leftUp[2]-Inf(Box[2]))*(rightUp[1]-Inf(Box[1]))-(leftDown[2]-Inf(Box[2]))*(rightDown[1]-Inf(Box[1])))/2;
    ivector temp(2);
    temp[1]=Sup(rightUp[1]);
    temp[2]=Sup(leftUp[2]);
    if(!triangle(1, rightUp, temp, leftUp,  h, i, Imono))
      return false;
    
    temp=Inf(Box);
    if(!triangle(3, rightDown, leftDown, temp,  h, i, Imono))
      return false;
    Box1[1]=interval(Sup(rightUp[1]), Sup(Box[1]));
    Box1[2]=interval(Inf(Box[2]), Sup(leftUp[2]));
    Box2[1]=Box[1];
    Box2[2]=interval(Sup(leftUp[2]), Sup(Box[2]));
  }

  interval htest=H(ivector(mid(Box1)));
  if(Inf(htest)>Sup(h)) {}
  else if(Sup(htest)<Inf(h))
    Integrate(i,  Imono, Box1);
  else
    return false;

  htest=H(ivector(mid(Box2)));
  if(Inf(htest)>Sup(h)) {}
  else if(Sup(htest)<Inf(h))
    Integrate(i,  Imono, Box2);
  else
    return false;
  
  ivector errBox=leftDown|leftUp|rightUp|rightDown;
  //  cout<<errArea<<endl;
  errArea=abs(errArea);

  Imono[i][1]+=interval(0.0)|errArea;
  Imono[i][2]+=interval(0.0)|errArea*sqr(errBox[1]);
  Imono[i][3]+=interval(0.0)|errArea*sqr(errBox[2]);
  Imono[i][4]+=interval(0.0)|errArea*power(errBox[1],4);
  Imono[i][5]+=interval(0.0)|errArea*sqr(errBox[1])*sqr(errBox[2]);
  Imono[i][6]+=interval(0.0)|errArea*power(errBox[2],4);

  return true;
}

void splitAndStore(const ivector &Box, List<ivector> *cover) {
  ivector temp(2);
  temp[1]=interval(Inf(Box[1]),mid(Box[1]));
  temp[2]=interval(Inf(Box[2]),mid(Box[2]));
  *cover+=temp;
  temp[2]=interval(mid(Box[2]),Sup(Box[2]));
  *cover+=temp;
  temp[1]=interval(mid(Box[1]),Sup(Box[1]));
  *cover+=temp;
  temp[2]=interval(Inf(Box[2]),mid(Box[2]));
  *cover+=temp;
}

interval I(const int &Index, const interval &h, const imatrix &Imono, const ivector &coef) {
  ivector currentImono=Imono[Index];
  interval retI(0.0);

  for(int j=1; j<=d; j++) {
    retI+=coef[j]*currentImono[j];
  }

  return retI;
}

interval lambda(const int &Index, const interval &h, const imatrix &Imono, const ivector &coef) {
  ivector currentImono=Imono[Index];
  if(Inf(currentImono[1])>0.0) {
    interval retI(0.0);
  
    for(int j=2; j<=d; j++) {
      retI+=coef[j]*currentImono[j];
    }
    
    return retI/currentImono[1];
  }
  return interval(-10000);
}


int main(int argc, char * argv[]) {
  if ( argc != 3 ) {
    cout << "Syntax: " << argv[0] << " tol1 oval " << endl;
    exit(0);
  }
  cout << endl
       << "**********************************************************" << endl
       << "******************* Imono - Integrator ********************" << endl
       << endl << "Latest compilation: " << __DATE__ << ", " << __TIME__
       << endl << endl;
  clock(START_TIMING);
  
  cout<< SetPrecision(7,6);
  //cout<< SetPrecision(17,16);
 
  oval=atoi(argv[2]);
  
  if (oval>9 || oval<1) {
    cout<<"Put 1<=oval <= 9 !"<<endl;
    exit(0);
  }

  d=6;
  
  real tol1(atof(argv[1]));
  real tol2(tol1/10.0);

  ivector coef(d);

  coef[1]= 2.176832745375219;
  coef[2]= 0.203687169951339;
  coef[3]=-4.663680776344302;
  coef[4]=-8.410822908376025;
  coef[5]= 4.313536179874701;
  coef[6]= 1.000000000000000;

  int sampleNr;
  rvector h(4);  

  cout<<coef<<endl;
  imatrix Imono(4,d); 
  
  if(oval==1) {
    sampleNr=3;
    h[1]=-1.8400;
    h[2]=-1.3800;
    h[3]=-1.3105;
  }
  else if(oval==2) {
    sampleNr=2;
    h[1]=-1.3456;
    h[2]=-1.3105;
  }
  else if(oval==3) {
    sampleNr=2;
    h[1]=-1.2673;
    h[2]=-0.7000;
  }
  else if(oval==6) {   //BACK, Witdraw 7 if h>0, i.e. h<0 with normalisation
    sampleNr=2;
    h[1]=-0.6000;
    h[2]= 0.0580;
    back=true;
  }   
  else if(oval==8) {   //BACK
    sampleNr=2;
    h[1]=0.0600;
    h[2]=0.1315;
    back=true;
  }
  else if(oval==9) {   //BACK
    sampleNr=4;
    h[1]=0.1341;
    h[2]=0.1413;
    h[3]=0.1637;
    h[4]=0.1861;
    back=true;
  }
  else
    exit(0);
  if(back)
    h=-h;


  ////////////////////////////
  for(int i=1; i<=sampleNr; i++) {
    for(int j=1; j<=6; j++) {
      Imono[i][j]=0;
    }
    
    List<ivector> cover;
    ivector Box(2);
    
    if(oval==1) {
      Box[1]=interval(0.5, 2.0);
      Box[2]=interval(0.75, 3.0);
    }
    else if(oval==2) {
      Box[1]=interval(-0.5, 0.5);
      Box[2]=interval(1.0, 3.0);
    }
    else if(oval==3) {
      Box[1]=interval(-2.0, 2.0);
      Box[2]=interval(0.75, 3.0);
    }
    else if(oval==4) {
      Box[1]=interval(1.0, 2.0);
      Box[2]=interval(-0.75, 0.75);
    }
    /*
    else if(oval==5) {   //Withdraw 6 from the result
      Box[1]=interval(-2.0, 2.0);
      Box[2]=interval(-3.0, 3.0);
    }
    */
    else if(oval==6) {   //BACK, Witdraw 7 if h>0, i.e. h<0 with normalisation
      Box[1]=interval(-1.5, 1.5);
      Box[2]=interval(-2.0, 2.0);
    }
    else if(oval==7) {
      Box[1]=interval(-0.5, 0.5);
      Box[2]=interval(-0.5, 0.5);
    }
    else if(oval==8) {   //BACK
      Box[1]=interval(-1.0, 1.0);
      Box[2]=interval(0.0, 1.5);
    }
    else if(oval==9) {   //BACK
      Box[1]=interval(0.0, 1.0);
      Box[2]=interval(0.0, 1.5);
    }
    else
      exit(0);
    
    cover+=Box;

    while(!IsEmpty(cover)) {
      Box=Pop(cover);
      interval Htest=H(Box);
      //cout<<Htest<<endl;
      if( Sup(Htest)<h[i] ) {
	Integrate(i,  Imono, Box);
      }
      else if(Inf(Htest)<=h[i] ){
	bool rotate=false;
	int p(0);
	real tempTol=tol2;
	ivector currImono=Imono[i];
	ivector orgImono=Imono[i];

	while(boundaryIntegrateRot(i,  interval(h[i]), Imono, Box, tempTol) && p<10) {
	  rotate=true;
	  p++;
	  tempTol/=2.0;
	  currImono=Imono[i];
	  Imono[i]=orgImono;
	}	  

	if(!rotate) {
	  if(max(diam(Box[1]),diam(Box[2]))<tol1) {
	    boundaryIntegrate(i, Imono, Box);
	    intNo++;
	  }      
	  else
	    splitAndStore(Box, &cover);
	}
	else {
	  Imono[i]=currImono;
	  intYes++;
	}
      }
      // else
// 	cout<<Box<<endl<<H(Box)<<endl<<endl;
    }

    if(oval==6 && h[i]<0.0) {
      h[i]=-h[i];
      //      Imono[i]=-Imono[i];
      back=false;
      Box[1]=interval(-0.5, 0.5);
      Box[2]=interval(-0.5, 0.5);
      cover+=Box;
      while(!IsEmpty(cover)) {
	Box=Pop(cover);
	interval Htest=H(Box);
	if( Sup(Htest)<h[i] ) {
	  Integrate(i,  Imono, Box);
	}
	else if(Inf(Htest)<=h[i] ){
	  bool rotate=false;
	  int p(0);
	  real tempTol=tol2;
	  ivector currImono=Imono[i];
	  ivector orgImono=Imono[i];
	  
	  while(boundaryIntegrateRot(i,  interval(h[i]), Imono, Box, tempTol) && p<10) {
	    rotate=true;
	    p++;
	    tempTol/=2.0;
	    currImono=Imono[i];
	    Imono[i]=orgImono;
	  }	  
	  
	  if(!rotate) {
	    if(max(diam(Box[1]),diam(Box[2]))<tol1) {
	      boundaryIntegrate(i, Imono, Box);
	      intNo++;
	    }      
	    else
	      splitAndStore(Box, &cover);
	  }
	  else {
	    Imono[i]=currImono;
	    intYes++;
	  }
	}
      }
      // Imono[i]=-Imono[i];
      h[i]=-h[i];
      back=true;
    }

    cout<<"Completed computations of Imono for Oval number "<<i<<endl;
    if(!back)
      cout<<"At the oval H = "<<h[i]<<endl;
    else
      cout<<"At the oval H = "<<-h[i]<<endl;

    cout<<"I(h) = "<<I(i, interval(h[i]), Imono, coef)<<endl;

  }

  cout<<endl<<"The number of boxes covering the oval is "<<intYes<<endl
      <<"The number of unidentifiable boxes is "<<intNo<<endl;
  clock(STOP_TIMING);
  //  cout<<Imono<<endl;
}
